"""
.. include:: ../../static/doc.md
"""
from typing import Sequence

from dali_renderer.dali import Dali
from dali_renderer.dali import ExistingPath
from dali_renderer.dali import CodeLength
from dali_renderer.dali import SyntaxNotFound
from dali_renderer.dali import PaddingError
from dali_renderer.dali import StyleNotFound
from dali_renderer.dali import InputNotSpecified


__all__: Sequence = [
    Dali,
    ExistingPath,
    CodeLength,
    SyntaxNotFound,
    PaddingError,
    StyleNotFound,
    InputNotSpecified,
]



import os
from typing import Union

from PIL import ImageFont

from dali_renderer.config import FONTS_PATH

STYLES = {
    "NORMAL": ["", "Regular", "Medium"],
    "ITALIC": ["Italic"],
    "BOLD": ["Bold", "SemiBold", "ExtraBold"],
    "BOLDITALIC": ["Bold Italic"],
}


class FontNotFound(Exception):
    """When there are no usable fonts specified"""


class DaliFontManager:
    """
    Manages a set of fonts: normal, italic, bold, etc...
    """

    def __init__(self, font_name: str, font_size: int = 14) -> None:
        self.font_name: str = font_name
        if font_size < 10:
            font_size = 14
        self.font_size: int = font_size
        self.fonts: dict = {}
        self._create()

    def _get_font_path(self, name: str, style_name: str) -> Union[str, None]:
        """
        Returning font path for style
        :param name: font name
        :type: str
        :param style_name: style name - const. STYLES
        :type: str
        :return: Full font path or None
        :rtype: Union[str, None]

        :raises: :class:`FontNotFound`: font not found or doesn't exist
        """
        font_dir_path = os.path.join(FONTS_PATH, name)
        try:
            for file in os.listdir(font_dir_path):
                font = ImageFont.truetype(os.path.join(font_dir_path, file))
                try:
                    font_name, font_style = font.getname()
                except OSError:
                    raise FontNotFound(f"No usable fonts named: {self.font_name}")
                if font_style == style_name:
                    font_path = os.path.join(font_dir_path, file)
                    if os.path.exists(font_path):
                        return font_path
            return None
        except FileNotFoundError:
            raise FontNotFound(f"No usable fonts named: {self.font_name}")

    def _create(self) -> None:
        """
        Dictionary generation with font style paths

        :return: None
        """
        for name in STYLES["NORMAL"]:
            path = self._get_font_path(self.font_name, name)
            if path is not None:
                self.fonts["NORMAL"] = ImageFont.truetype(path, self.font_size)
                break
        else:
            raise FontNotFound(f"No usable fonts named: {self.font_name}")
        for style in ("ITALIC", "BOLD", "BOLDITALIC"):
            for stylename in STYLES[style]:
                path = self._get_font_path(self.font_name, stylename)
                if path is not None:
                    self.fonts[style] = ImageFont.truetype(path, self.font_size)
                    break
            else:
                if style == "BOLDITALIC":
                    self.fonts[style] = self.fonts["BOLD"]
                else:
                    self.fonts[style] = self.fonts["NORMAL"]

    def get_char_size(self) -> tuple:
        """
        Get the character size.
        """
        return self.get_text_size("M")

    def get_text_size(self, text: str) -> tuple:
        """
        Get the text size (width, height).
        """
        font = self.fonts["NORMAL"]
        if hasattr(font, "getbbox"):  # Pillow >= 9.2.0
            return font.getbbox(text)[2:4]
        else:
            return font.getsize(text)

    def get_font(self, bold: bool, oblique: bool) -> ImageFont.FreeTypeFont:
        """
        Get the font based on bold and italic flags.
        """
        if bold and oblique:
            return self.fonts["BOLDITALIC"]
        elif bold:
            return self.fonts["BOLD"]
        elif oblique:
            return self.fonts["ITALIC"]
        else:
            return self.fonts["NORMAL"]



import os
import tempfile
import shutil
from os import PathLike
from pathlib import Path
from typing import KeysView, Union

from pygments.lexers import guess_lexer, get_all_lexers, get_lexer_by_name
from pygments.lexer import Lexer
from pygments import highlight
from pygments.styles import STYLE_MAP, get_style_by_name
from PIL import Image
from cairosvg import svg2png

from dali_renderer.renders.svg import SVGUtils
from dali_renderer.renders.DaliImageFormatter import ImageFormatter

from dali_renderer.config import BACKGROUND_WRAPPER
from dali_renderer.config import TMP_PNG_WRAPPER_NAME
from dali_renderer.config import TMP_SVG_WRAPPER_NAME
from dali_renderer.config import TMP_IMAGE_NAME
from dali_renderer.config import CONTENT_FILE_NAME
from dali_renderer.config import DEFAULT_FONT


class ExistingPath(Exception):
    pass


class CodeLength(Exception):
    pass


class SyntaxNotFound(Exception):
    pass


class InputNotSpecified(Exception):
    pass


class PaddingError(Exception):
    pass


class StyleNotFound(Exception):
    pass


class Dali:
    """
    Main `Dali` class
    :param output_path: Path to save the renderer image
    :type output_path: str
    :param syntax: Language lexer ex. `python`
    :type syntax: str
    :param style: Highlight style, default `one-dark`
    :type style: str
    :param font: One of the fonts `Fira, Hack, Inconsolata, JetBrains`, default `Hack`
    :type font: str
    :param padding: Image padding (if you need to render a small code-snippet,
        set the padding to less than or equal to 30), default `50`
    :type padding: int
    :param font_size: Font size in pixels, default `21`
    :type font_size: int
    :param border_radius: Border radius in pixels, default `10`
    :type border_radius: int
    :param window_controls: Set if you need to set window decoration on your image
    :type window_controls: bool
    :param background: Set if you need to set additional background around of the image
    :type background: str
    """

    output_extension = [".png"]

    def __init__(
        self,
        output_path: str,
        syntax: str = "",
        style: str = "one-dark",
        font: str = DEFAULT_FONT,
        padding: int = 50,
        font_size: int = 21,
        border_radius: int = 10,  # set 0 to disable,
        window_controls: bool = False,
        background: str = "",
    ):
        self._input_file: Union[PathLike, str] = Path()
        self.output_path: str = output_path
        self.syntax: str = syntax
        self.style: str = style
        self.font_name: str = font
        self.padding: int = padding
        self.font_size: int = font_size
        self.border_radius: int = border_radius
        self.window_controls: bool = window_controls
        self.background: str = background

        # If we are drawing window controls, we should use padding equal to 50 or more
        if self.window_controls and self.padding < 50:
            self.padding = 50

        self._tmp_dir = tempfile.mkdtemp()
        self._tmp_svg_wrapper = os.path.join(self._tmp_dir, TMP_SVG_WRAPPER_NAME)
        self._tmp_output_png_wrapper = os.path.join(self._tmp_dir, TMP_PNG_WRAPPER_NAME)
        self._tmp_output_image = os.path.join(self._tmp_dir, TMP_IMAGE_NAME)
        self._tmp_content_file = os.path.join(self._tmp_dir, CONTENT_FILE_NAME)

        self._height: int = 0
        self._width: int = 0

        if self.padding < 10:
            raise PaddingError(
                f"Padding cannot be less than 10. Given value is {self.padding}"
            )
        elif self.padding > 100:
            raise PaddingError(
                f"Padding cannot be more than 10. Given value is {self.padding}"
            )

        if not self._check_output_extension():
            raise InputNotSpecified(
                f"The given output filename `{self.output_path}` doesn't have a valid extension"
            )

    @property
    def styles(self) -> KeysView[str]:
        return STYLE_MAP.keys()

    @staticmethod
    def get_supported_lexers(only_name: bool = False) -> list:
        """
        Return all available lexers
        :param only_name: return only lower case lexer names
        :type only_name: bool
        :return: list of lexers
        :rtype: list
        """
        lexers = [
            lexer[0].lower() if only_name else lexer for lexer in get_all_lexers()
        ]

        return lexers

    def from_string(self, code: str) -> None:
        """
        Generation an image from a string of source code
        :param code: string of source code
        :type code: str
        :return: None
        """
        if len(code) <= 10:
            raise CodeLength(
                "Code length must be a positive integer and greater than 10"
            )

        self._set_file_content(code)
        self._input_file = self._tmp_content_file
        self._generate()

    def from_file(self, file_path: str) -> None:
        """
        Generation an image from a source code file
        :param file_path: the path to the source code file
        :type file_path: str
        :return: None
        """
        if not os.path.exists(file_path):
            raise ExistingPath(f"Input path {file_path} doesn't exist")

        if os.path.getsize(file_path) <= 10:
            raise CodeLength(
                "Code length must be a positive integer and greater than 10"
            )

        self._input_file = file_path
        self._generate()

    def _generate(self) -> str:
        """
        The basic method of manipulating the generated image
        :return: absolute path to the generated image
        :rtype: str
        """
        if self.syntax and not self._is_supported_syntax():
            raise SyntaxNotFound(f"`{self.syntax}` syntax is not supported")

        if self.style and not self._is_supported_style():
            raise StyleNotFound(f"`{self.style}` style is not supported")

        formatter: ImageFormatter = ImageFormatter(
            font_name=self.font_name,
            font_size=self.font_size,
            style=self.style,
            line_number_fg=None,
        )

        """Forming a basic highlighted image"""
        self.get_supported_lexers()
        with open(self._input_file, "r") as file_input:
            content = file_input.read()
            with open(self._tmp_output_image, "wb") as file_output:
                """lexer definition"""
                lexer: Lexer
                if self.syntax:
                    lexer = get_lexer_by_name(self.syntax)
                else:
                    lexer = guess_lexer(content)
                file_output.write(highlight(content, lexer, formatter))

        basic_image = Image.open(self._tmp_output_image)
        self._width, self._height = basic_image.size

        """Forming the svg wrapper"""
        self.svg: SVGUtils = SVGUtils(
            BACKGROUND_WRAPPER, get_style_by_name(self.style).background_color
        )
        self.svg.set_width(self._width + (self.padding * 2))
        self.svg.set_height(self._height + (self.padding * 2))
        if self.border_radius:
            self.svg.set_rounded_corners(self.border_radius)

        """SVG wrapper to PNG"""
        if self.background:
            self.svg.set_additional_background(self.background)
            self._width = self._width + self.svg.background_padding
            self._height = self._height + self.svg.background_padding
            self.padding = self.padding + (self.svg.background_padding // 2)

        if self.window_controls:
            self.svg.draw_controls()

        self.svg.save(self._tmp_svg_wrapper)

        svg2png(
            url=self._tmp_svg_wrapper,
            write_to=self._tmp_output_png_wrapper,
            parent_width=self.svg.width,
            parent_height=self.svg.height,
        )

        """Moving image to wrapper"""
        background_im = Image.open(self._tmp_output_png_wrapper)
        background_im.paste(basic_image, (self.padding, self.padding), basic_image)

        background_im.save(self.output_path, compress_level=0)

        return self.output_path

    def _is_supported_syntax(self) -> bool:
        """
        Syntax support check
        :return: check result
        :rtype: bool
        """
        if self.syntax not in self.get_supported_lexers(True):
            return False

        return True

    def _is_supported_style(self) -> bool:
        """
        Style support check
        :return: check result
        :rtype: bool
        """
        if self.style not in self.styles:
            return False

        return True

    def _set_file_content(self, content: str) -> None:
        """
        Save the content to a file
        :param content: source code content
        :type content: str
        :return: None
        """
        with open(self._tmp_content_file, "w") as file:
            file.write(content)

    def _check_output_extension(self) -> bool:
        """
        Output extension checking
        :return: check result
        :rtype: bool
        """
        for ext in self.output_extension:
            if self.output_path.lower().endswith(ext):
                return True
        return False

    def __del__(self) -> None:
        """
        Delete all temporary files
        :return: None
        """
        shutil.rmtree(self._tmp_dir)



import os

SOURCE_DIR = os.path.dirname(__file__)
DEFAULT_FONT = "Hack"
TMP_SVG_WRAPPER_NAME = "dali_wrapper.svg"
TMP_PNG_WRAPPER_NAME = "dali_wrapper.png"
TMP_IMAGE_NAME = "dali_renderer.png"
ASSETS_DIR = os.path.join(SOURCE_DIR, "assets")
CONTENT_FILE_NAME = "content"

FONTS_PATH = os.path.join(ASSETS_DIR, "fonts")
BACKGROUND_WRAPPER = os.path.join(ASSETS_DIR, "background_wrapper.svg")

"""test defs"""
TESTS_PATH = os.path.abspath("tests")



from typing import Tuple, Union, Iterator

try:
    from PIL import Image, ImageDraw, ImageFont  # noqa

    pil_available = True
except ImportError:
    pil_available = False
from pygments.formatter import Formatter
from pygments.util import get_bool_opt, get_int_opt, get_list_opt, get_choice_opt

from dali_renderer.renders.DaliFontManager import DaliFontManager


class PilNotAvailable(ImportError):
    """When Python imaging library is not available"""


class ImageFormatter(Formatter):
    """
    Create a PNG image from source code. This uses the Python Imaging Library to
    generate a pixmap from the source code.

    .. versionadded:: 0.10

    Additional options accepted:

    `image_format`
        An image format to output to that is recognised by PIL, these include:

        * "PNG" (default)
        * "JPEG"
        * "BMP"
        * "GIF"

    `line_pad`
        The extra spacing (in pixels) between each line of text.

        Default: 2

    `font_name`
        The font name to be used as the base font from which others, such as
        bold and italic fonts will be generated.  This really should be a
        monospace font to look sane.

        Default: "Hack"

    `font_size`
        The font size in points to be used.

        Default: 14

    `image_pad`
        The padding, in pixels to be used at each edge of the resulting image.

        Default: 10

    `line_numbers`
        Whether line numbers should be shown: True/False

        Default: True

    `line_number_start`
        The line number of the first line.

        Default: 1

    `line_number_step`
        The step used when printing line numbers.

        Default: 1

    `line_number_bg`
        The background colour (in "#123456" format) of the line number bar, or
        None to use the style background color.

        Default: "#eed"

    `line_number_fg`
        The text color of the line numbers (in "#123456"-like format).

        Default: "#886"

    `line_number_chars`
        The number of columns of line numbers allowable in the line number
        margin.

        Default: 2

    `line_number_bold`
        Whether line numbers will be bold: True/False

        Default: False

    `line_number_italic`
        Whether line numbers will be italicized: True/False

        Default: False

    `line_number_separator`
        Whether a line will be drawn between the line number area and the
        source code area: True/False

        Default: True

    `line_number_pad`
        The horizontal padding (in pixels) between the line number margin, and
        the source code area.

        Default: 6

    `hl_lines`
        Specify a list of lines to be highlighted.

        .. versionadded:: 1.2

        Default: empty list

    `hl_color`
        Specify the color for highlighting lines.

        .. versionadded:: 1.2

        Default: highlight color of the selected style

    `rounded`
        Radius of rounded corners. 0 to disable

        Default: 10
    """

    # Required by the pygments mapper
    name = "img"
    aliases = ["img", "IMG", "png"]
    filenames = ["*.png"]

    unicodeoutput = False

    default_image_format = "png"

    def __init__(self, **options: Union[str, int, None]):
        """
        See the class docstring for explanation of options.
        """
        if not pil_available:
            raise PilNotAvailable(
                "Python Imaging Library is required for this formatter"
            )
        Formatter.__init__(self, **options)  # type: ignore
        self.encoding: str = "latin1"  # let pygments.format() do the right thing
        # Read the style
        self.styles: dict = dict(self.style)
        self.background_color: str
        if self.style.background_color is None:
            self.background_color = "#fff"
        else:
            self.background_color = self.style.background_color
        # Image options
        self.image_format: str = get_choice_opt(
            options,
            "image_format",
            ["png", "jpeg", "gif", "bmp"],
            self.default_image_format,
            normcase=True,
        )
        self.image_pad: int = get_int_opt(options, "image_pad", 10)
        self.line_pad: int = get_int_opt(options, "line_pad", 2)
        # The fonts
        self.fontsize: int = get_int_opt(options, "font_size", 14)
        self.fonts: DaliFontManager = DaliFontManager(options.get("font_name", ""), self.fontsize)  # type: ignore
        # noinspection PyTupleAssignmentBalance
        self.fontw, self.fonth = self.fonts.get_char_size()
        # Line number options
        self.line_number_fg: str = options.get("line_number_fg", "#886")  # type: ignore
        self.line_number_bg: str = options.get("line_number_bg", "#eed")  # type: ignore
        self.line_number_chars: int = get_int_opt(options, "line_number_chars", 2)
        self.line_number_bold: bool = get_bool_opt(options, "line_number_bold", False)
        self.line_number_italic: bool = get_bool_opt(
            options, "line_number_italic", False
        )
        self.line_number_pad: int = get_int_opt(options, "line_number_pad", 6)
        self.line_numbers: bool = get_bool_opt(options, "line_numbers", True)
        self.line_number_separator: bool = get_bool_opt(
            options, "line_number_separator", True
        )
        self.line_number_step: int = get_int_opt(options, "line_number_step", 1)
        self.line_number_start: int = get_int_opt(options, "line_number_start", 1)
        self.line_number_width: int
        if self.line_numbers:
            self.line_number_width = (
                self.fontw * self.line_number_chars + self.line_number_pad * 2
            )
        else:
            self.line_number_width = 0
        self.hl_lines: list = []
        hl_lines_str: list = get_list_opt(options, "hl_lines", [])
        for line in hl_lines_str:
            try:
                self.hl_lines.append(int(line))
            except ValueError:
                pass
        self.hl_color: str = options.get("hl_color", self.style.highlight_color) or "#f90"  # type: ignore
        self.drawables: list = []

    def get_style_defs(self, arg: str = "") -> None:
        raise NotImplementedError(
            "The -S option is meaningless for the image "
            "formatter. Use -O style=<stylename> instead."
        )

    def _get_line_height(self) -> int:
        """
        Get the height of a line.
        """
        return self.fonth + self.line_pad

    def _get_line_y(self, lineno: int) -> int:
        """
        Get the Y coordinate of a line number.
        """
        return lineno * self._get_line_height() + self.image_pad

    def _get_char_width(self) -> int:
        """
        Get the width of a character.
        """
        return self.fontw

    def _get_char_x(self, linelength: int) -> int:
        """
        Get the X coordinate of a character position.
        """
        return linelength + self.image_pad + self.line_number_width

    def _get_text_pos(self, linelength: int, lineno: int) -> Tuple[int, int]:
        """
        Get the actual position for a character and line position.
        """
        return self._get_char_x(linelength), self._get_line_y(lineno)

    def _get_linenumber_pos(self, lineno: int) -> Tuple[int, int]:
        """
        Get the actual position for the start of a line number.
        """
        return self.image_pad, self._get_line_y(lineno)

    def _get_text_color(self, style: dict) -> str:
        """
        Get the correct color for the token from the style.
        """
        if style["color"] is not None:
            fill = "#" + style["color"]
        else:
            fill = "#000"
        return fill

    def _get_text_bg_color(self, style: dict) -> str:
        """
        Get the correct background color for the token from the style.
        """
        if style["bgcolor"] is not None:
            bg_color = "#" + style["bgcolor"]
        else:
            bg_color = None
        return bg_color

    def _get_style_font(self, style: dict) -> ImageFont.FreeTypeFont:
        """
        Get the correct font for the style.
        """
        return self.fonts.get_font(style["bold"], style["italic"])

    def _get_image_size(self, maxlinelength: int, maxlineno: int) -> Tuple[int, int]:
        """
        Get the required image size.
        """
        return (
            self._get_char_x(maxlinelength) + self.image_pad,
            self._get_line_y(maxlineno + 0) + self.image_pad,
        )

    def _draw_linenumber(self, posno: int, lineno: int) -> None:
        """
        Remember a line number drawable to paint later.
        """
        self._draw_text(
            self._get_linenumber_pos(posno),
            str(lineno).rjust(self.line_number_chars),
            font=self.fonts.get_font(self.line_number_bold, self.line_number_italic),
            text_fg=self.line_number_fg,
            text_bg=None,
        )

    def _draw_text(
        self,
        pos: Tuple[int, int],
        text: str,
        font: ImageFont.FreeTypeFont,
        text_fg: str,
        text_bg: Union[str, None],
    ) -> None:
        self.drawables.append((pos, text, font, text_fg, text_bg))

    def _create_drawables(self, tokensource: Iterator) -> None:
        """
        Create drawables for the token content.
        """
        lineno: int
        charno: int
        maxcharno: int
        lineno = charno = maxcharno = 0
        maxlinelength: int
        linelength: int
        maxlinelength = linelength = 0
        for ttype, value in tokensource:
            while ttype not in self.styles:
                ttype = ttype.parent
            style = self.styles[ttype]
            # TODO: make sure tab expansion happens earlier in the chain.  It
            # really ought to be done on the input, as to do it right here is
            # quite complex.
            value = value.expandtabs(4)
            lines = value.splitlines(True)
            # print lines
            for i, line in enumerate(lines):
                temp = line.rstrip("\n")
                if temp:
                    self._draw_text(
                        self._get_text_pos(linelength, lineno),
                        temp,
                        font=self._get_style_font(style),
                        text_fg=self._get_text_color(style),
                        text_bg=self._get_text_bg_color(style),
                    )
                    # noinspection PyTupleAssignmentBalance
                    temp_width, _ = self.fonts.get_text_size(temp)
                    linelength += temp_width
                    maxlinelength = max(maxlinelength, linelength)
                    charno += len(temp)
                    maxcharno = max(maxcharno, charno)
                if line.endswith("\n"):
                    # add a line for each extra line in the value
                    linelength = 0
                    charno = 0
                    lineno += 1
        self.maxlinelength = maxlinelength
        self.maxcharno = maxcharno
        self.maxlineno = lineno

    def _draw_line_numbers(self) -> None:
        """
        Create drawables for the line numbers.
        """
        if not self.line_numbers:
            return
        for p in range(self.maxlineno):
            n = p + self.line_number_start
            if (n % self.line_number_step) == 0:
                self._draw_linenumber(p, n)

    def _paint_line_number_bg(self, im: Image.Image) -> None:
        """
        Paint the line number background on the image.
        """
        if not self.line_numbers:
            return
        if self.line_number_fg is None:
            return
        draw: ImageDraw.ImageDraw = ImageDraw.Draw(im)
        recth: int = im.size[-1]
        rectw: int = self.image_pad + self.line_number_width - self.line_number_pad
        draw.rectangle([(0, 0), (rectw, recth)], fill=self.line_number_bg)  # type: ignore
        draw.line([(rectw, 0), (rectw, recth)], fill=self.line_number_fg)
        del draw

    def format(self, tokensource: Iterator, outfile: str) -> None:
        """
        Format ``tokensource``, an iterable of ``(tokentype, tokenstring)``
        tuples and write it into ``outfile``.

        This implementation calculates where it should draw each token on the
        pixmap, then calculates the required pixmap size and draws the items.
        """
        self._create_drawables(tokensource)
        self._draw_line_numbers()
        width: int
        height: int
        width, height = self._get_image_size(self.maxlinelength, self.maxlineno)

        im: Image.Image = Image.new(
            "RGBA",
            (width, height),
            self.background_color,
        )

        draw: ImageDraw.ImageDraw = ImageDraw.Draw(im)

        # self._paint_line_number_bg(im)

        # Highlight
        if self.hl_lines:
            x = self.image_pad + self.line_number_width - self.line_number_pad + 1
            recth: int = self._get_line_height()
            rectw: int = im.size[0] - x
            for linenumber in self.hl_lines:
                y = self._get_line_y(linenumber - 1)
                draw.rectangle([(x, y), (x + rectw, y + recth)], fill=self.hl_color)  # type: ignore
        for pos, value, font, text_fg, text_bg in self.drawables:
            if text_bg:
                text_size: Tuple[int, int] = draw.textsize(text=value, font=font)
                # noinspection PyTypeChecker
                draw.rectangle(
                    [pos[0], pos[1], pos[0] + text_size[0], pos[1] + text_size[1]],  # type: ignore
                    fill=text_bg,
                )

            draw.text((pos[0], pos[1]), value, font=font, fill=text_fg)

        im.save(outfile, self.image_format.upper(), quality=100)



import re
from os import PathLike
from typing import Union

from bs4 import BeautifulSoup
from bs4.element import Tag, NavigableString


class SizeError(Exception):
    pass


class UnknownColor(Exception):
    pass


class SVGUtils:
    def __init__(self, filepath: str, background: str) -> None:
        self.filepath: str = filepath

        self._additional_background: Union[str, None] = None

        self._is_valid_color(background)
        self.background: str = background

        with open(self.filepath, "r") as svg_file:
            self._soup: BeautifulSoup = BeautifulSoup(svg_file.read(), "xml")

        self._svg_tag: Union[Tag, NavigableString, None] = self._soup.find("svg")
        self._rect: Union[Tag, NavigableString, None] = self._soup.find("rect")
        self._g: Union[Tag, NavigableString, None] = self._soup.find("g")

        self._height: int = 0
        self._width: int = 0

        self._background_padding: int = 200

        self._set_background_color()

    @property
    def background_padding(self) -> int:
        """
        Get background width padding
        :return: width integer
        :rtype: int
        """
        return self._background_padding

    @property
    def height(self) -> int:
        """
        Get SVG height
        :return: height integer
        :rtype: int
        """
        return self._height

    @property
    def width(self) -> int:
        """
        Get SVG width
        :return: width integer
        :rtype: int
        """
        return self._width

    @property
    def get_content(self) -> str:
        """
        Get the contents of the `BeautifulSoup` object as a string
        :return: file content
        :rtype: str
        """
        return str(self._soup)

    @property
    def get_object(self) -> BeautifulSoup:
        """
        Get the `BeautifulSoup` object
        :return:
        """
        return self._soup

    def set_height(self, height: int) -> None:
        """
        Setting the height of the main rectangle
        :param height: height in pixels
        :type height: int
        :return: None
        """
        if height <= 0:
            raise SizeError(
                f"{height} must be a positive integer and greater than zero"
            )

        self._height = height
        self._rect.attrs["height"] = height  # type: ignore

    def set_width(self, width: int) -> None:
        """
        Setting the width of the main rectangle
        :param width: width in pixels
        :type width: int
        :return: None
        """
        if width <= 0:
            raise SizeError(f"{width} must be a positive integer and greater than zero")

        self._width = width
        self._rect.attrs["width"] = width  # type: ignore

    def set_rounded_corners(self, radius: int = 10) -> None:
        """
        Setting rounded corners of the main rectangle
        :param radius: radius in pixels
        :type radius: int
        :return: None
        """
        if radius < 0:
            raise SizeError(f"{radius} must be a positive integer or zero")

        self._rect.attrs["rx"] = radius  # type: ignore

    def set_additional_background(self, color: str) -> None:
        """
        Setting an additional background
        :return: None
        """
        self._is_valid_color(color)
        self._additional_background = color

        if not self._width or not self._height:
            raise SizeError(
                "You have to set height and width first. Use `set_height` and `set_width`"
            )

        background_tag: Tag = self._soup.new_tag(
            "rect",
            attrs={
                "style": f"fill:{color}",
                "height": (self._height + self._background_padding),
                "width": (self._width + self._background_padding),
            },
        )

        self._g.insert(0, background_tag)  # type: ignore

        self._rect.attrs["y"] = self._background_padding // 2  # type: ignore
        self._rect.attrs["x"] = self._background_padding // 2  # type: ignore
        self._svg_tag.attrs["height"] = self._height + self._background_padding  # type: ignore
        self._svg_tag.attrs["width"] = self._width + self._background_padding  # type: ignore

    def draw_controls(self) -> None:
        """
        Drawing window controls
        :return: None
        """
        close_control = self._soup.new_tag(
            "circle",
            attrs={
                "cx": (35 + (self._background_padding // 2))
                if self._additional_background
                else 35,
                "cy": (
                    32 + (self._background_padding // 2)
                    if self._additional_background
                    else 32
                ),
                "r": 10,
                "fill": "#ff5d57",
            },
        )
        hide_control = self._soup.new_tag(
            "circle",
            attrs={
                "cx": (70 + (self._background_padding // 2))
                if self._additional_background
                else 70,
                "cy": (
                    32 + (self._background_padding // 2)
                    if self._additional_background
                    else 32
                ),
                "r": 10,
                "fill": "#ffbc30",
            },
        )
        maximize_control = self._soup.new_tag(
            "circle",
            attrs={
                "cx": (105 + (self._background_padding // 2))
                if self._additional_background
                else 105,
                "cy": (
                    32 + (self._background_padding // 2)
                    if self._additional_background
                    else 32
                ),
                "r": 10,
                "fill": "#27c93f",
            },
        )

        self._rect.insert_after(close_control)  # type: ignore
        close_control.insert_before(hide_control)
        hide_control.insert_before(maximize_control)

    def save(self, path: Union[PathLike, str]) -> None:
        """
        Saving the edited file
        :param path: path to save temporary svg file
        :type path: Union[PathLike, str]
        :return: None
        """

        """Saving edited svg"""
        with open(path, "w") as file:
            file.write(self.get_content)

    def _set_background_color(self) -> None:
        self._rect.attrs["style"] = f"fill: {self.background}"  # type: ignore

    def _is_valid_color(self, color: str) -> bool:
        """
        Validate color hex string
        :param color:
        :return:
        """
        if not re.match(r"^#(?:[0-9a-fA-F]{3}){1,2}$", color):
            raise UnknownColor(
                f"Please, set a valid color in hex format (Value: {color})"
            )

        return True
